import os
import re
import logging
from itertools import chain

import numpy as np
from astropy.io import fits
from astropy.table import Table, vstack
from astropy.time import Time, TimeDelta
from astropy.coordinates import SkyCoord
from . import archiveutils as au

"""
Represents a VLBA UVFITS file
"""

LOG = logging.getLogger('ingestion')

class UVFits:
    def __init__(self, file):
        self.fitsfile = file

        with fits.open(self.fitsfile) as fh:
            self.source_table = Table(fh['SOURCE'].data)
            self.source_header = fh['SOURCE'].header
            self.reference_frequency = fh['FREQUENCY'].header['REF_FREQ']
            self.frequency_table = Table(fh['FREQUENCY'].data)
            self.uv_header = fh['UV_DATA'].header
            self.scans = Table()

            i = 1
            start_scan = 0
            # Read through trying to find the UV_DATA tables (there may be multiple in older idifits file)
            # Scan number should continue to increment across UV_DATA tables
            while True:
                try:
                    if fh[i].header['EXTNAME'] == 'UV_DATA':
                        uv_data = fh[i].data
                        start_scan, table_scans = self.build_scans(uv_data, start_scan)
                        self.scans = vstack([self.scans, table_scans])
                    i = i + 1
                except OSError:
                    break

    def getsource(self, source_id):
        return self.source_table[self.source_table['ID_NO.'] == source_id]['SOURCE']

    def build_scans(self, uv_data, start_scan):
        scan_list = []
        polarizations = self.get_polarizations()

        source = uv_data[0]['SOURCE']
        scan = start_scan
        scan_start_row = 0
        frequency_ids = [uv_data[scan_start_row]['FREQID']]
        for row in range(len(uv_data)):
            if source != uv_data[row]['SOURCE']:
                scan_end_row = row - 1
                scan_start_time = Time(uv_data[scan_start_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_start_row]['TIME'], format='jd')
                scan_end_time = Time(uv_data[scan_end_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_end_row]['TIME'], format='jd')
                scan_time_on_source = scan_end_time - scan_start_time
                scan_integration_time = uv_data[scan_start_row]['INTTIM']
                scan_source = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['SOURCE'][0]
                source_ra = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['RAAPP'][0]
                source_dec = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['DECAPP'][0]
                c = SkyCoord(source_ra, source_dec, frame='icrs', unit='deg')
                source_radec = c.to_string('hmsdms')
                frequencies = [(self.get_reference_frequency() + self.getfrequency(f)) for f in frequency_ids]
                frequency_list = np.concatenate(frequencies, axis=0) / 1000000
                total_bandwidth = self.getbandwidth(1)
                current_scan = [scan, scan_source, scan_start_time.mjd, scan_end_time.mjd, scan_time_on_source.sec, scan_integration_time, ' '.join(map(str, frequency_list)), total_bandwidth, source_ra, source_dec, source_radec, polarizations]
                scan_list.append(current_scan)
                scan = scan + 1
                scan_start_row = row
                frequency_ids.clear()

            frequency_id = uv_data[row]['FREQID']
            if frequency_id not in frequency_ids:
                frequency_ids.append(frequency_id)

            source = uv_data[row]['SOURCE']

        # Add in final scan
        scan_end_row = row
        scan_start_time = Time(uv_data[scan_start_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_start_row]['TIME'], format='jd')
        scan_end_time = Time(uv_data[scan_end_row]['DATE'], format='jd') + TimeDelta(uv_data[scan_end_row]['TIME'], format='jd')
        scan_time_on_source = scan_end_time - scan_start_time
        scan_integration_time = uv_data[scan_start_row]['INTTIM']
        scan_source = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['SOURCE'][0]
        source_ra = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['RAAPP'][0]
        source_dec = self.source_table[self.source_table['ID_NO.'] == uv_data[scan_start_row]['SOURCE']]['DECAPP'][0]
        c = SkyCoord(source_ra, source_dec, frame='icrs', unit='deg')
        source_radec = c.to_string('hmsdms')
        frequencies = [(self.get_reference_frequency() + self.getfrequency(f)) for f in frequency_ids]
        frequency_list = np.concatenate(frequencies, axis=0) / 1000000
        total_bandwidth = self.getbandwidth(1)
        current_scan = [scan, scan_source, scan_start_time.mjd, scan_end_time.mjd, scan_time_on_source.sec, scan_integration_time, ' '.join(map(str, frequency_list)), total_bandwidth, source_ra, source_dec, source_radec, polarizations]
        scan_list.append(current_scan)

        # Build scan table from list of scans and increment start scan number in case there is another UV_DATA table
        return scan + 1, Table(rows=scan_list, names=('SCAN_ID', 'SOURCE', 'START_TIME', 'END_TIME', 'TIME_ON_SOURCE', 'INTEGRATION_TIME', 'FREQUENCY', 'BANDWIDTH', 'RA', 'DEC', 'RA/DEC', 'POLARIZATION'))

    # Observation attributes

    def getfitsfile(self):
        return self.fitsfile

    def number_of_scans(self):
        return len(self.scans)

    def get_reference_frequency(self):
        return self.reference_frequency

    # Scan attributes

    def get_source(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'SOURCE')

    def get_start_time(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'START_TIME')

    def get_end_time(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'END_TIME')

    def get_ra(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'RA')

    def get_dec(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'DEC')

    def get_integration_time(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'INTEGRATION_TIME')

    def get_bandwidth(self, scan_id):
        return self.get_column_scalar_data(scan_id, 'BANDWIDTH')

    # Helper function

    def get_column_scalar_data(self, scan_id, column_name):
        return self.scans[self.scans['SCAN_ID'] == scan_id][column_name].data[0]

    # Polarization

    def get_polarizations(self):
        # The final polarization code is the sum of the individual ones
        polarization_code = 0
        number_polarizations = self.uv_header['NO_STKD']
        for i in range(number_polarizations):
            pol_key = f"STK_{i+1}"
            if pol_key not in self.uv_header.keys():
                LOG.warning(f"Stokes key: {pol_key} is not present in UV_DATA table")
                continue
            polarization = self.lookup_polarization(self.uv_header[pol_key])
            polarization_code = polarization_code + polarization
        return au.getpolarization(polarization_code)
        return polarization_code

    def lookup_polarization(self, polarization_code):
        # Map of polarization codes used by FITS to observation polarization codes
        polarization_codes = {
            1:  'I',
            2:  'Q',
            3:  'U',
            4:  'V',
            -1: 'RR',
            -2: 'LL',
            -3: 'RL',
            -4: 'LR',
            -5: 'XX',
            -6: 'YY',
            -7: 'XY',
            -8: 'YX'
        }

        if polarization_code not in polarization_codes.keys():
            LOG.error(f"Invalid polarization code: {polarization_code}")
            exit(-5)

        return au.getpolarizationcode(polarization_codes[polarization_code])

    def getproject(self):
        return self.source_header['OBSCODE']

    def getexperiment(self):
        return ''

    def getcorrelatorpass(self):
        fflist = os.path.basename(self.getfitsfile()).split('_')
        return fflist[2]

    def getobservingband(self):
        ref_frequency = self.reference_frequency
        return self.lookup_observing_band(ref_frequency)

    def getfrequency(self, frequency):
        frequency_row = self.frequency_table[self.frequency_table['FREQID'] == frequency]
        frequencies = frequency_row[0]['BANDFREQ']
        if np.isscalar(frequencies):
            # If the frequency is a scalar, put it in a numpy ndarray
            freq_array = np.ndarray(shape=(1))
            freq_array[0] = frequency_row[0]['BANDFREQ']
            return freq_array
        return frequencies

    def getbandwidth(self, frequency):
        frequency_row = self.frequency_table[self.frequency_table['FREQID'] == frequency]
        if np.isscalar(frequency_row[0]['TOTAL_BANDWIDTH']):
            return frequency_row[0]['TOTAL_BANDWIDTH']
        else:
            return frequency_row[0]['TOTAL_BANDWIDTH'][0]

    def lookup_observing_band(self, reference_frequency):
        ref_freq = reference_frequency / 1000000000
        obs_band = ''

        if 0.058 <= ref_freq <= 0.084:
            obs_band = '4'
        elif 0.236 <= ref_freq <= 0.492:
            obs_band = 'P'
        elif 1.0 <= ref_freq <= 2.0:
            obs_band = 'L'
        elif 2.0 < ref_freq <= 4.0:
            obs_band = 'S'
        elif 4.0 < ref_freq <= 8.0:
            obs_band = 'C'
        elif 8.0 < ref_freq <= 12.0:
            obs_band = 'X'
        elif 12.0 < ref_freq <= 18.0:
            obs_band = 'Ku'
        elif 18.0 < ref_freq <= 26.5:
            obs_band = 'K'
        elif 26.5 < ref_freq <= 40.0:
            obs_band = 'Ka'
        elif 40.0 < ref_freq <= 50.0:
            obs_band = 'Q'
        else:
            LOG.warning(f"Reference frequency {ref_freq} could not be located in a band")

        return obs_band

    def dump_scans(self):
        self.scans.pprint(max_lines=-1, max_width=-1)

def test_file(file):
    if os.path.isfile(file):
        print(f"File {file} exists")
        assert True
    else:
        print(f"File {file} doesn't exist'")
        assert False


def test_headers(file):
    # with fits.open(file) as fh:
    #     for i in range(1, 17):
    #         print(fh[i].header['EXTNAME'])
    pass


def test_source(uv_fits, source_id, expected):
    source_name = uv_fits.getsource(source_id)
    assert source_name == expected


def test_scans(uv_fits):
    columns = uv_fits.scans.columns
    num_rows = len(uv_fits.scans[:])
    print(f"Number of rows: {num_rows}")
    LOG.error(columns)


if __name__ == '__main__':
    path = '/Users/rlively/vlba'
    file = 'VLBA_VSN005405_file24.uvfits'

    uv_fits_file = os.path.join(path, file)
    uv_fits = UVFits(uv_fits_file)

    print(f"Fits file: {uv_fits.get_fits_file()}")
    print(uv_fits.get_polarizations())
    uv_fits.scans.pprint(max_lines=-1, max_width=-1)
    print(uv_fits.get_polarizations())


